#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

#define BLOCK_SIZE 256
#define UNROLL_FACTOR 4

// Kernel to reduce warp divergence by using predicated execution
// and minimizing conditional branches within warps

template <typename scalar_t, int CHANNELS>
__global__ void warp_divergence_reduced_softmax_sigmoid_kernel(
    const scalar_t* __restrict__ input,
    scalar_t* __restrict__ output,
    const int batch,
    const int depth,
    const int height,
    const int width) {

    const int spatial = depth * height * width;
    const int total_pixels = batch * spatial;
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < total_pixels) {
        const int b = idx / spatial;
        const int pixel_idx = idx % spatial;
        const int d = pixel_idx / (height * width);
        const int rem = pixel_idx % (height * width);
        const int h = rem / width;
        const int w = rem % width;

        const int base = (b * CHANNELS * spatial) + (d * height * width + h * width + w);
        const int stride = spatial;

        scalar_t max_val = input[base];
        scalar_t sum_exp = 1.0f;

        #pragma unroll
        for (int c = 0; c < CHANNELS; c += UNROLL_FACTOR) {
            scalar_t vals[UNROLL_FACTOR];
            #pragma unroll
            for (int u = 0; u < UNROLL_FACTOR && (c + u) < CHANNELS; ++u) {
                const int pos = base + (c + u) * stride;
                vals[u] = input[pos];
            }
            #pragma unroll
            for (int u = 0; u < UNROLL_FACTOR && (c + u) < CHANNELS; ++u) {
                scalar_t val = vals[u];
                bool update_max = val > max_val;
                max_val = update_max ? val : max_val;
                sum_exp = update_max ? sum_exp * exp(max_val - val) + 1.0f : sum_exp + exp(val - max_val);
            }
        }

        #pragma unroll
        for (int c = 0; c < CHANNELS; c += UNROLL_FACTOR) {
            #pragma unroll
            for (int u = 0; u < UNROLL_FACTOR && (c + u) < CHANNELS; ++u) {
                const int pos = base + (c + u) * stride;
                const scalar_t softmax_val = exp(input[pos] - max_val) / sum_exp;
                output[pos] = 1.0f / (1.0f + exp(-softmax_val));
            }
        }
    }
}

// Non-templated version for dynamic channel sizes
template <typename scalar_t>
__global__ void dynamic_warp_divergence_reduced_softmax_sigmoid_kernel(
    const scalar_t* __restrict__ input,
    scalar_t* __restrict__ output,
    const int batch,
    const int channels,
    const int depth,
    const int height,
    const int width) {

    const int spatial = depth * height * width;
    const int total_pixels = batch * spatial;
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;

    if (idx < total_pixels) {
        const int b = idx / spatial;
        const int pixel_idx = idx % spatial;
        const int d = pixel_idx / (height * width);
        const int rem = pixel_idx % (height * width);
        const int h = rem / width;
        const int w = rem % width;

        const int base = (b * channels * spatial) + (d * height * width + h * width + w);
        const int stride = spatial;

        scalar_t max_val = input[base];
        scalar_t sum_exp = 1.0f;

        #pragma unroll 4
        for (int c = 0; c < channels; ++c) {
            const int pos = base + c * stride;
            const scalar_t val = input[pos];
            bool update_max = val > max_val;
            max_val = update_max ? val : max_val;
            sum_exp = update_max ? sum_exp * exp(max_val - val) + 1.0f : sum_exp + exp(val - max_val);
        }

        #pragma unroll 4
        for (int c = 0; c < channels; ++c) {
            const int pos = base + c * stride;
            const scalar_t softmax_val = exp(input[pos] - max_val) / sum_exp;
            output[pos] = 1.0f / (1.0f + exp(-softmax_val));
        }
    }
}

torch::Tensor forward(
    torch::Tensor input,
    int stride,
    int padding,
    int output_padding,
    bool bias_flag,
    torch::Tensor conv_transpose,
    torch::Tensor conv_transpose_bias) {

    auto x = torch::conv_transpose3d(
        input,
        conv_transpose,
        bias_flag ? conv_transpose_bias : torch::Tensor(),
        stride,
        padding,
        output_padding
    );

    const int batch = x.size(0);
    const int channels = x.size(1);
    const int depth = x.size(2);
    const int height = x.size(3);
    const int width = x.size(4);

    auto output = torch::empty_like(x);

    const int spatial = depth * height * width;
    const int total_pixels = batch * spatial;
    const int threads = BLOCK_SIZE;
    const int blocks = (total_pixels + threads - 1) / threads;

    AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "warp_divergence_reduced_softmax_sigmoid_kernel", ([&] {
        // Choose specialized kernel based on channel count
        if (channels == 32) {
            warp_divergence_reduced_softmax_sigmoid_kernel<scalar_t, 32><<<blocks, threads>>>(
                x.data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch,
                depth,
                height,
                width);
        } else if (channels == 64) {
            warp_divergence_reduced_softmax_sigmoid_kernel<scalar_t, 64><<<blocks, threads>>>(
                x.data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch,
                depth,
                height,
                width);
        } else {
            dynamic_warp_divergence_reduced_softmax_sigmoid_kernel<scalar_t><<<blocks, threads>>>(
                x.data_ptr<scalar_t>(),
                output.data_ptr<scalar_t>(),
                batch,
                channels,
                depth,
                height,
                width);
        }
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Warp Divergence Reduced Fused ConvTranspose3d with Softmax and Sigmoid");
}
